/**
* cmd_handler.c - Executes iptables shell commands 
* sent by the access point module via a message queue
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

//mqueue specific decls
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"
#include <signal.h>
#include <errno.h>

key_t mq_key;
int mq_id;
#define MTYPE_FW_CMD 10 //message type: itpables command (the only valid mtype)
#define MAX_MSG_LEN 1024

typedef struct msgbuf {
	long mtype;
	char mtext[1];
} MSG_BUF;

//initializes message queue
int init_mq() {
	mq_id = msgget(mq_key, IPC_CREAT | 0400);
	if (mq_id == -1) {
		fprintf(stderr, "init_mq: Error getting message queue\n");
		fprintf(stderr, "%s  -  %d\n", strerror(errno), errno);
	}
	return mq_id;
}

//sends command to message queue to be processed by the other process
int rcv_cmd() {
	//FW_CMD cmd;
	MSG_BUF *msg_to_receive;

	ALLOC_AND_CLEAN(msg_to_receive, MAX_MSG_LEN, MSG_BUF*); //allocate memory for msgbuf

	if (msgrcv(mq_id, msg_to_receive, MAX_MSG_LEN, MTYPE_FW_CMD, 0) == -1) {
		fprintf(stderr, "rcv_cmd: Error receiving message from queue\n");
		return -1;
	}
	else {
		system(msg_to_receive->mtext);
	}
	return 0;
}

//removes message queue
int remove_mq() {
	int retval = 0;
	int qid = msgget(mq_key, 0600);
	retval = msgctl(qid, IPC_RMID, NULL);
	if (retval == -1) {
		fprintf(stderr, "remove_mq: Error removing message queue\n");
	}
	return retval;
}

//################################################################

void sighandler(int signum) {
	switch (signum) {
		case SIGINT: case SIGTERM: case SIGKILL:
		remove_mq();
		return;
	}
}

int main(int argc, char **argv) {
	signal(SIGINT, sighandler);
	signal(SIGTERM, sighandler);

	if (argc < 2) {
		fprintf(stderr, "Message queue key not specified\n");
		exit(1);
	}

	mq_key = atoi(argv[1]);

	init_mq();
	while (!rcv_cmd()) {
	}
	return 0;
}
