/**
* tlog.c - Traffic logging kernel module
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/socket.h>
#include <linux/netdevice.h>

#include <linux/string.h>
#include <linux/types.h>
#include <linux/net.h>
#include <linux/skbuff.h>
#include <asm/errno.h>
#include <asm/uaccess.h>

#include <linux/ip.h>

#define MAX_CLIENTS 1024


MODULE_LICENSE("GPL");

MODULE_AUTHOR("Pantelis Frangoudis, Elias Efstathiou");

MODULE_DESCRIPTION("Netfilter-based traffic logging module for the P2PWNC access point server");

MODULE_SUPPORTED_DEVICE("tstats");

static struct nf_hook_ops nfho;
static struct nf_hook_ops nfho2;

/**
 * statistics structs
 */
struct stats {
	unsigned int nClientIP;
	unsigned int nTotalPacketsIn;
	unsigned int nTotalPacketsOut;
	unsigned int nTotalBytesIn;
	unsigned int nTotalBytesOut;
};

struct client_list {
	unsigned int nRecNum;
	struct stats clist[MAX_CLIENTS];
};


/**
 * ioctl codes
 */
#define ACTIVATE   0xFEED0000
#define DEACTIVATE   0xFEED0001
#define GET_STATS   0xFEED0002
#define REMOVE_REC 0xFEED0003
#define ADD_REC 0xFEED0004
#define MAJOR_NUM 241


static int active = 0;
static int stats_ctrl_in_use = 0;
static int major = 0;
//static struct stats netstats = {0, 0, 0};
static struct client_list clients;
static unsigned int recnum = 0;

static int stats_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int stats_open(struct inode *inode, struct file *file);
static int stats_release(struct inode *inode, struct file *file);


struct file_operations stats_fops = {
	.owner = THIS_MODULE,
	.ioctl = stats_ioctl,
	.open = stats_open,
	.release = stats_release
};

static int stats_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg) {
	int i = 0;
	int j = 0;
	switch (cmd) {
		case ACTIVATE:
			active = 1;
			printk("Traffic statistics module activated.\n");
			return 0;

		case DEACTIVATE:
			active ^= active;
			printk("Traffic statistics module deactivated.\n");
			return 0;

		case GET_STATS:
			//returns stats for the specified ip
			//printk("Request for traffic stats\n");

			if ((struct stats *)arg == NULL) {
				printk("Traffic statistics module: NULL argument\n");
				return -EINVAL;
			}
			for (i=0;i<clients.nRecNum;i++) {
				if (clients.clist[i].nClientIP == *(unsigned int*)arg) {
					copy_to_user((struct stats*)arg, &(clients.clist[i]), sizeof(struct stats));
					//printk( "Traffic: %d\n", ((struct stats*)arg)->nTotalBytesIn );
					break;
				}
			}
			return 0;

		case REMOVE_REC:
			//arg: ip diey8ynsh pou prepei na diagrafei
			//psaxnei sth lista, th diagrafei kai kanei shift back ta ypoloipa....
			for (i=0;i<clients.nRecNum;i++) {
				if (clients.clist[i].nClientIP == *(unsigned int*)arg) {
					for (j=i; j<clients.nRecNum-1; j++) {
						clients.clist[j].nClientIP = clients.clist[j+1].nClientIP;
						clients.clist[j].nTotalPacketsIn = clients.clist[j+1].nTotalPacketsIn;
						clients.clist[j].nTotalPacketsOut = clients.clist[j+1].nTotalPacketsOut;
						clients.clist[j].nTotalBytesIn = clients.clist[j+1].nTotalBytesIn;
						clients.clist[j].nTotalBytesOut = clients.clist[j+1].nTotalBytesOut;
					}
					clients.clist[recnum-1].nClientIP = 0;
					clients.clist[recnum-1].nTotalPacketsIn = 0;
					clients.clist[recnum-1].nTotalPacketsOut = 0;
					clients.clist[recnum-1].nTotalBytesIn = 0;
					clients.clist[recnum-1].nTotalBytesOut = 0;

					clients.nRecNum--;
					recnum--;
					break;
				}
			}
			return 0;

		case ADD_REC:
			//arg: ip diey8ynsh poy 8a proste8ei
			for (i=0;i<clients.nRecNum;i++) {
				if (clients.clist[i].nClientIP == *(unsigned int*)arg) {
					break;
				}
			}
			if (i == clients.nRecNum) {
				//nea eggrafh - den yphrxe sth lista
				clients.clist[clients.nRecNum].nClientIP = *(unsigned int*)arg;
				clients.clist[i].nTotalPacketsIn = 0;
				clients.clist[i].nTotalPacketsOut = 0;
				clients.clist[i].nTotalBytesIn = 0;
				clients.clist[i].nTotalBytesOut = 0;
				clients.nRecNum++;
				recnum++;
				printk("Traffic stats module: New record added\n");
			}
			else {
				printk("Traffic stats module: Record already exists\n");
			}
			return 0;

		default:
			return -EBADRQC;
	}
}

static int stats_open(struct inode *inode, struct file *file)
{
	if (stats_ctrl_in_use) {
		return -EBUSY;
	}
	else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		try_module_get(THIS_MODULE);
#else
		MOD_INC_USE_COUNT;
#endif
		stats_ctrl_in_use++;
		return 0;
	}
	return 0;
}

static int stats_release(struct inode *inode, struct file *file) {
	stats_ctrl_in_use ^= stats_ctrl_in_use;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	module_put(THIS_MODULE);
#else
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

/**
 * netfilter hook function - measurements take place here
 */
unsigned int hook_func(unsigned int hooknum,
			struct sk_buff **skb,
			const struct net_device *in,
			const struct net_device *out,
			int (*okfn)(struct sk_buff *))
{

	//if module is not active, accept all traffic
	//and do nothing
	if (!active) {
		return NF_ACCEPT;
	}

	unsigned int ipsrc = (*skb)->nh.iph->saddr;
	unsigned int ipdst = (*skb)->nh.iph->daddr;

	int i = 0;
	//match ip
	int srcfound = -1;
	int dstfound = -1;
	for (i=0; i<recnum; i++) {
		if ( clients.clist[i].nClientIP == ipsrc) {
			srcfound = i;
		}
		if (clients.clist[i].nClientIP == ipdst){
			dstfound = i;
		}
	}

	if (srcfound != -1) {
		clients.clist[srcfound].nTotalPacketsOut++;
		clients.clist[srcfound].nTotalBytesOut += (*skb)->len;
	}
	if (dstfound != -1) {
		clients.clist[dstfound].nTotalPacketsIn++;
		clients.clist[dstfound].nTotalBytesIn += (*skb)->len;
	}

	return NF_ACCEPT;
}

/**
 * module_init routine
 */
static int __init tlog_init(void) {
	SET_MODULE_OWNER(&stats_fops);
	if ((major = register_chrdev(MAJOR_NUM, "tstats",
				&stats_fops)) < 0) {
		printk("Traffic Statistics Module: Failed registering control device!\n");
		printk("Traffic Statistics Module: Module installation aborted.\n");
		return major;
	}
	else {
		printk("Traffic Statistics Module: Register Char Dev: OK\n");
	}

	stats_ctrl_in_use ^= stats_ctrl_in_use;
	active = 1;
	printk("Traffic statistics module: Success!\n");
	printk("Traffic statistics module activated.\n");
	clients.nRecNum = 0;
	recnum = 0;

	nfho.hook     = hook_func;
	nfho.hooknum  = NF_IP_PRE_ROUTING;
	nfho.pf       = PF_INET;
	nfho.priority = NF_IP_PRI_FIRST;
	nf_register_hook(&nfho);

	nfho2.hook     = hook_func;
	nfho2.hooknum  = NF_IP_POST_ROUTING;
	nfho2.pf       = PF_INET;
	nfho2.priority = NF_IP_PRI_FIRST;
	nf_register_hook(&nfho2);

	return 0;
}

/**
 * module_exit (cleanup) routine
 */
static void __exit tlog_cleanup(void) {
	nf_unregister_hook(&nfho);
	nf_unregister_hook(&nfho2);
	if (unregister_chrdev(MAJOR_NUM, "tstats") != 0) {
		printk("Traffic Statistics Module: Removal of module failed!\n");
	}
	else {
		printk("Traffic Statistics Module: Removal of module suceeded!\n");
	}
}

module_init(tlog_init);
module_exit(tlog_cleanup);
